/*
 * ThinkRequest.java
 *
 * Created on 3. mai 2007, 11:44
 *
 * This class represents a session of the game ThinkAlike, it is created when a user
 * wants to start a session, and when a user recieves a challange.
 *
 */

package thinkAlike;

import java.io.IOException;

/**
 * This class represents a session of the game ThinkAlike, it is created when a user wants to start a session, and when a user recieves a challange.
 * @author Runar Gunnerud
 * @version 1.0
 */
public class ThinkRequest {
    
    private ProgramControl pc;
    private String to, from, pic, thinkword, word1, word2, word3;
    private boolean isInitiator = false;
    
    /**
     * Creates a new instance of ThinkRequest
     * @param pc pointer to the ProgramControl class
     * @param to the address of the recipient of the initial session
     * @param from the address of the initiator of the session
     * @param pic a text representation of the picture, encoded in Base64
     * @param thinkword the word the initator assosiates with the picture
     */
    public ThinkRequest(ProgramControl pc, String to, String from, String pic, String thinkword) {
        this.pc = pc;
        this.to = to;
        this.from = from;
        this.pic = pic;
        this.thinkword = thinkword;
    }
    
    /**
     * This method is called when the a ThinkRequest is recieved, showing a screen asking the user to confirm if he want the session or not.
     */
    public void confirmRequest(){
        pc.getSc().get_fromUserString().setText(this.from);
        pc.getSc().getDisplay().setCurrent(pc.getSc().get_recievedChallengeForm());
        pc.getSc().getDisplay().vibrate(2000);
        pc.getSc().getDisplay().flashBacklight(3000);
    }
    

    /**
     * deprecated
     */
    public void requestConfirmed(){
    }
    
    /**
     * Sends the actual request using the Jxa object. It does not send the object itself, but the picture and the thinkword to the recipient.
     * @throws java.io.IOException On communication errors from the Jxa object.
     */
    public void sendThinkRequest() throws IOException{
        pc.getJxa().sendThinkRequest(this.to, this.from, this.pic, this.thinkword);
    }
    
    /**
     * When recieveing an answer to the challenge, this message get the three words suggested and puts them on the screen of the initiator. 
     * It also uses the calculateResult() to figure out if the recipient though of the same word.
     * @param word1 first word
     * @param word2 second word
     * @param word3 third word
     */
    public void recievedResponse(String word1, String word2, String word3){
        setWords(word1, word2, word3);
        pc.getSc().get_word1Recieved().setLabel("First word:");
        pc.getSc().get_word1Recieved().setText(word1);
        pc.getSc().get_word2Recieved().setLabel("Second word:");
        pc.getSc().get_word2Recieved().setText(word2);
        pc.getSc().get_word3Recieved().setLabel("Third word:");
        pc.getSc().get_word3Recieved().setText(word3);
        
        String resultTxt = calculateResult();
        
        pc.getSc().get_resultRequestString().setText(resultTxt);
    }
    
    /**
     * The session at the recipient, this method generates the form that show the result of her guessing.
     */
    public void makeResultForm(){
        String resultTxt = calculateResult();
        pc.getSc().get_theWordRecieverString().setText(thinkword);
        pc.getSc().get_suggestedWordsRecieverString().setText(word1 + " " + word2 + " " + word3);
        pc.getSc().get_resultRecieverString().setText(resultTxt);
    }
    
    /**
     * On both part of the session this method generates the end form where the participans also can chat.
     */
    public void makeEndForm(){
        String resultTxt = calculateResult();
        pc.getSc().get_initialWordString().setText(thinkword);
        pc.getSc().get_suggestedWordsEndString().setText(word1 + " " + word2 + " " + word3);
        
    }
    
    /**
     * Simply checks if one of the three suggested word of the recipient matches the initial word and returns a string matching the result.
     * @return a string that describes the outcome of the check
     */
    public String calculateResult() {
        if(thinkword.equals(word1) || thinkword.equals(word2) || thinkword.equals(word3)){
            return "Yo you rock! and ThinkAlike!";
        } else {
            return "You're soooo far off!";
        }
    }
    

    /**
     * Used by the recipient to send the reply of the ThinkRequest to the initiator, using the Jxa object to send the three words. 
     * 
     * Notice that the response is sent back to the requester, therefore the to field of the message is the from variable in the ThinkRequest calss.
     * @throws java.io.IOException in case of communication error when using the Jxa object
     */
    public void sendWords() throws IOException{
        pc.getJxa().sendThinkRespond(this.from, this.to, this.word1, this.word2, this.word3);
    }
    
    /**
     * Sets if this ThinkRequest object conatins the initiator or not (the other option is obviously the recipient)
     * @param is is true if it is the initiator
     */
    public void isInitiator(boolean is){
        this.isInitiator = is;
    }
    
    /**
     * Returns if this ThinkRequest object is acting as an initiator or not
     * @return true if object is initiator, false if recipient
     */
    public boolean isInitiator(){
        return this.isInitiator;
    }
    
    /**
     * Simple method to set the three words suggested by the recipient
     * @param word1 first word
     * @param word2 second word
     * @param word3 third word
     */
    public void setWords(String word1, String word2, String word3){
        this.word1 = word1;
        this.word2 = word2;
        this.word3 = word3;
    }
    
    /**
     * Returns the the address of the recipient fo the session
     * @return address of recipient
     */
    public String getTo(){
        return to;
    }
    /**
     * Returns the the address of the initiator fo the session
     * @return address of initiator
     */
    public String getFrom(){
        return from;
    }
    /**
     * Returns the the picture assosiated with this session, as a text string in Base64 encoding
     * @return picture as text string
     */
    public String getPic(){
        return pic;
    }
    /**
     * Returns the the word that the initiator assosiated with the picture
     * @return TheWord
     */
    public String getThinkWord(){
        return thinkword;
    }
    /**
     * Return the first word suggested by the recipient
     * @return first word
     */
    public String getWord1(){
        return word1;
    }
    /**
     * Return the second word suggested by the recipient
     * @return second word
     */
    public String getWord2(){
        return word2;
    }
    /**
     * Return the third word suggested by the recipient
     * @return third word
     */
    public String getWord3(){
        return word3;
    }
    
    
    
}
