/*
 * Copyright 2004-2006 Swen Kummer, Dustin Hass, Sven Jost, Grzegorz Grasza
 * http://jxa.sourceforge.net/
 *
 * This is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version. Mobber is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details. You should have received a copy of the GNU General Public License
 * along with mobber; if not, write to the Free Software Foundation, Inc., 59
 * Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 * The tow methods sendThinkRquest and sendThinkResponse is added by Runar Gunnerud
 * for use in the ThinkAlike application, it is released under the same license.
 *
 */

package net.sourceforge.jxa;

import javax.microedition.io.*;
import java.io.*;
import java.util.*;

/**
 * J2ME XMPP API Class
 *
 * @author Swen Kummer, Dustin Hass, Sven Jost, Grzegorz Grasza
 * @version 4.0
 * @since 1.0
 */

public class Jxa extends Thread {
    
    private final boolean DEBUG = true;
    
    private final String host, port, username, resource, password, myjid;
    private final int priority;
    
    private XmlReader reader;
    private XmlWriter writer;
    
    private Vector listeners = new Vector();
    
    /**
     * If you create this object all variables will be saved and the
     * method {@link #run()} is started to log in on jabber server and
     * listen to parse incomming xml stanzas. Use
     * {@link #addListener(XmppListener xl)} to listen to events of this object.
     *
     * @param host the hostname/ip of the jabber server
     * @param port the port number of the jabber server
     * @param username the username of the jabber account
     * @param password the passwort of the jabber account
     * @param resource a unique identifier of the used resource, for e.g. "mobile"
     * @param priority the priority of the jabber session, defines on which
     * resource the messages arrive
     */
    public Jxa(final String host, final String port, final String username, final String password, final String resource, final int priority) {
        this.host = host;
        this.port = port;
        this.username = username;
        this.password = password;
        this.resource = resource;
        this.priority = priority;
        this.myjid = username + "@" + host;
        this.start();
    }
    
    /**
     * The <code>run</code> method is called when {@link Jxa} object is
     * created. It sets up the reader and writer, calls {@link #login()}
     * methode and listens on the reader to parse incomming xml stanzas.
     */
    public void run() {
        try {
            final StreamConnection connection = (StreamConnection) Connector.open("socket://" + this.host + ":" + this.port);
            this.reader = new XmlReader(connection.openInputStream());
            this.writer = new XmlWriter(connection.openOutputStream());
        } catch (final Exception e) {
            this.connectionFailed();
            return;
        }
        // connected
        try {
            this.login();
            this.parse();
        } catch (final Exception e) {
            try {
                e.printStackTrace();
                this.writer.close();
                this.reader.close();
            } catch (final IOException io) {
                // io.printStackTrace();
            }
            // e.printStackTrace();
            // hier entsteht der connection failed bug (Network Down)
            this.connectionFailed();
        }
    }
    
    /**
     * Add a {@link XmppListener} to listen for events.
     *
     * @param xl a XmppListener object
     */
    public void addListener(final XmppListener xl) {
        if(!listeners.contains(xl)) listeners.addElement(xl);
    }
    
    /**
     * Remove a {@link XmppListener} from this class.
     *
     * @param xl a XmppListener object
     */
    public void removeListener(final XmppListener xl) {
        listeners.removeElement(xl);
    }
    
    /**
     * Opens the connection with a stream-tag, queries authentication type and
     * sends authentication data, which is username, password and resource.
     *
     * @throws java.io.IOException is thrown if {@link XmlReader} or {@link XmlWriter}
     *	throw an IOException.
     */
    public void login() throws IOException {
        // start stream
        this.writer.startTag("stream:stream");
        this.writer.attribute("to", this.host);
        this.writer.attribute("xmlns", "jabber:client");
        this.writer.attribute("xmlns:stream", "http://etherx.jabber.org/streams");
        this.writer.flush();
        // log in
        this.writer.startTag("iq");
        this.writer.attribute("type", "set");
        this.writer.attribute("id", "auth");
        this.writer.startTag("query");
        this.writer.attribute("xmlns", "jabber:iq:auth");
        
        this.writer.startTag("username");
        this.writer.text(this.username);
        this.writer.endTag();
        this.writer.startTag("password");
        this.writer.text(this.password);
        this.writer.endTag();
        this.writer.startTag("resource");
        this.writer.text(this.resource);
        this.writer.endTag();
        
        this.writer.endTag(); // query
        this.writer.endTag(); // iq
        this.writer.flush();
    }
    
    /**
     * Closes the stream-tag and the {@link XmlWriter}.
     */
    public void logoff() {
        try {
            this.writer.endTag();
            this.writer.flush();
            this.writer.close();
        } catch (final Exception e) {
            this.connectionFailed();
        }
    }
    
    /**
     * Sends a message text to a known jid.
     *
     * @param to the JID of the recipient
     * @param msg the message itself
     */
    public void sendMessage(final String to, final String msg) {
        try {
            this.writer.startTag("message");
            this.writer.attribute("type", "chat");
            this.writer.attribute("to", to);
            this.writer.startTag("body");
            this.writer.text(msg);
            this.writer.endTag();
            this.writer.endTag();
            this.writer.flush();
        } catch (final Exception e) {
            // e.printStackTrace();
            this.connectionFailed();
        }
    }
    
    /**
     * Sends a presence stanza to a jid. This method can do various task but
     * it's private, please use setStatus to set your status or explicit
     * subscription methods subscribe, unsubscribe, subscribed and
     * unsubscribed to change subscriptions.
     */
    private void sendPresence(final String to, final String type, final String show, final String status, final int priority) {
        try {
            this.writer.startTag("presence");
            if (type != null) {
                this.writer.attribute("type", type);
            }
            if (to != null) {
                this.writer.attribute("to", to);
            }
            if (show != null) {
                this.writer.startTag("show");
                this.writer.text(show);
                this.writer.endTag();
            }
            if (status != null) {
                this.writer.startTag("status");
                this.writer.text(status);
                this.writer.endTag();
            }
            if (priority != 0) {
                this.writer.startTag("priority");
                this.writer.text(Integer.toString(priority));
                this.writer.endTag();
            }
            this.writer.endTag(); // presence
            this.writer.flush();
        } catch (final Exception e) {
            // e.printStackTrace();
            this.connectionFailed();
        }
    }
    
    /**
     * Sets your Jabber Status.
     *
     * @param show is one of the following: <code>null</code>, chat, away,
     *        dnd, xa, invisible
     * @param status an extended text describing the actual status
     * @param priority the priority number (5 should be default)
     */
    public void setStatus(String show, String status, final int priority) {
        if (show.equals("")) {
            show = null;
        }
        if (status.equals("")) {
            status = null;
        }
        if (show.equals("invisible")) {
            this.sendPresence(null, "invisible", null, null, priority);
        } else {
            this.sendPresence(null, null, show, status, priority);
        }
    }
    
    /**
     * Requesting a subscription.
     *
     * @param to the jid you want to subscribe
     */
    public void subscribe(final String to) {
        this.sendPresence(to, "subscribe", null, null, 0);
    }
    
    /**
     * Remove a subscription.
     *
     * @param to the jid you want to remove your subscription
     */
    public void unsubscribe(final String to) {
        this.sendPresence(to, "unsubscribe", null, null, 0);
    }
    
    /**
     * Approve a subscription request.
     *
     * @param to the jid that sent you a subscription request
     */
    public void subscribed(final String to) {
        this.sendPresence(to, "subscribed", null, null, 0);
    }
    
    /**
     * Refuse/Reject a subscription request.
     *
     * @param to the jid that sent you a subscription request
     */
    public void unsubscribed(final String to) {
        this.sendPresence(to, "unsubscribed", null, null, 0);
    }
    
    /**
     * Save a contact to roster. This means, a message is send to jabber
     * server (which hosts your roster) to update the roster.
     *
     * @param jid the jid of the contact
     * @param name the nickname of the contact
     * @param group the group of the contact
     * @param subscription the subscription of the contact
     */
    public void saveContact(final String jid, final String name, final Enumeration group, final String subscription) {
        try {
            this.writer.startTag("iq");
            this.writer.attribute("type", "set");
            this.writer.startTag("query");
            this.writer.attribute("xmlns", "jabber:iq:roster");
            this.writer.startTag("item");
            this.writer.attribute("jid", jid);
            if (name != null) {
                this.writer.attribute("name", name);
            }
            if (subscription != null) {
                this.writer.attribute("subscription", subscription);
            }
            if (group != null) {
                while (group.hasMoreElements()) {
                    this.writer.startTag("group");
                    this.writer.text((String) group.nextElement());
                    this.writer.endTag(); // group
                }
            }
            this.writer.endTag(); // item
            this.writer.endTag(); // query
            this.writer.endTag(); // iq
            this.writer.flush();
        } catch (final Exception e) {
            // e.printStackTrace();
            this.connectionFailed();
        }
    }
    
    /**
     * Sends a roster query.
     *
     * @throws java.io.IOException is thrown if {@link XmlReader} or {@link XmlWriter}
     *	throw an IOException.
     */
    public void getRoster() throws IOException {
        this.writer.startTag("iq");
        this.writer.attribute("id", "roster");
        this.writer.attribute("type", "get");
        this.writer.startTag("query");
        this.writer.attribute("xmlns", "jabber:iq:roster");
        this.writer.endTag(); // query
        this.writer.endTag(); // iq
        this.writer.flush();
    }
    
    /**
     * The main parse methode is parsing all types of XML stanzas
     * <code>message</code>, <code>presence</code> and <code>iq</code>.
     * Although ignores any other type of xml.
     *
     * @throws java.io.IOException is thrown if {@link XmlReader} or {@link XmlWriter}
     *	throw an IOException.
     */
    private void parse() throws IOException {
        if (DEBUG) java.lang.System.out.println("*debug* parsing");
        this.reader.next(); // start tag
        if (DEBUG) java.lang.System.out.println("*debug* parsing etter next()");
        while (this.reader.next() == XmlReader.START_TAG) {
            final String tmp = this.reader.getName();
            if (tmp.equals("message")) {
                this.parseMessage();
            } else if (tmp.equals("presence")) {
                this.parsePresence();
            } else if (tmp.equals("iq")) {
                this.parseIq();
            } else {
                this.parseIgnore();
            }
        }
        this.reader.close();
    }
    
    /**
     * This method parses all info/query stanzas, including authentication
     * mechanism and roster. It also answers version queries.
     *
     * @throws java.io.IOException is thrown if {@link XmlReader} or {@link XmlWriter}
     *	throw an IOException.
     *
     *  This method has been modified by Runar Gunnerud to suite the ThinkAlike applicaiton
     *  that uses a set of info/quesry stanzas to control programflow
     *
     */
    private void parseIq() throws IOException {
        if (DEBUG) java.lang.System.out.println("*debug* parsing IQ");
        String type = this.reader.getAttribute("type");
        final String id = this.reader.getAttribute("id");
        final String from = this.reader.getAttribute("from");
        if (type.equals("error")) {
            while (this.reader.next() == XmlReader.START_TAG) {
                // String name = reader.getName();
                if (this.reader.getName().equals("error")) {
                    final String code = this.reader.getAttribute("code");
                    for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
                        XmppListener xl = (XmppListener) e.nextElement();
                        xl.onAuthFailed(code + ": " + this.parseText());
                    }
                } else {
                    this.parseText();
                }
            }
        } else if (type.equals("result") && (id != null) && id.equals("auth")) {
            // authorized
            this.reader.next();
            for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
                XmppListener xl = (XmppListener) e.nextElement();
                xl.onAuth();
            }
            this.sendPresence(null, null, null, null, this.priority);
        } else {
            while (this.reader.next() == XmlReader.START_TAG) {
                if (this.reader.getName().equals("query")) {
                    if (this.reader.getAttribute("xmlns").equals("jabber:iq:roster")) {
                        while (this.reader.next() == XmlReader.START_TAG) {
                            if (this.reader.getName().equals("item")) {
                                type = this.reader.getAttribute("type");
                                String jid = reader.getAttribute("jid"),
                                        name = reader.getAttribute("name"),
                                        subscription = reader.getAttribute("subscription"),
                                        newjid = (jid.indexOf('/') == -1) ? jid : jid.substring(0, jid.indexOf('/'));
                                boolean check = true;
                                for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
                                    XmppListener xl = (XmppListener) e.nextElement();
                                    xl.onContactRemoveEvent(newjid);
                                }
                                while (this.reader.next() == XmlReader.START_TAG) {
                                    if (this.reader.getName().equals("group")) {
                                        for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
                                            XmppListener xl = (XmppListener) e.nextElement();
                                            xl.onContactEvent(newjid, name, this.parseText(), subscription);
                                        }
                                        check = false;
                                    } else {
                                        this.parseIgnore();
                                    }
                                }
                                if (check && !subscription.equals("remove")) {
                                    for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
                                        XmppListener xl = (XmppListener) e.nextElement();
                                        xl.onContactEvent(jid, name, "", subscription);
                                    }
                                }
                            } else {
                                this.parseIgnore();
                            }
                        }
                    } else if (this.reader.getAttribute("xmlns").equals("jabber:iq:version")) {
                        while (this.reader.next() == XmlReader.START_TAG) {
                            this.parseIgnore();
                        }
                        // reader.next();
                        // send version
                        this.writer.startTag("iq");
                        this.writer.attribute("type", "result");
                        this.writer.attribute("id", id);
                        this.writer.attribute("to", from);
                        this.writer.startTag("query");
                        this.writer.attribute("xmlns", "jabber:iq:version");
                        
                        this.writer.startTag("name");
                        this.writer.text("jxa");
                        this.writer.endTag();
                        this.writer.startTag("version");
                        writer.text("1.0");
                        this.writer.endTag();
                        this.writer.startTag("os");
                        this.writer.text("J2ME");
                        this.writer.endTag();
                        
                        this.writer.endTag(); // query
                        this.writer.endTag(); // iq
                    }
                                        /*
                                         *  The part underneath is the extention made by Runar Gunnerud
                                         *  to handle messages from a ThinkAlike client.
                                         *
                                         *
                                         */
                    
                    else if (this.reader.getAttribute("xmlns").equals("com.runarg.thinkalike")) {
                        String thinktype = null, thinkword = null, pic = null, word1 = null, word2 = null, word3 = null;
                        while (this.reader.next() == XmlReader.START_TAG) {
                            final String tmp = this.reader.getName();
                            if (tmp.equals("thinktype")) {
                                thinktype = this.parseText();
                            } else if (tmp.equals("pic")) {
                                pic = this.parseText();
                            } else if (tmp.equals("thinkword")) {
                                thinkword = this.parseText();
                            } else if (tmp.equals("word1")) {
                                word1 = this.parseText();
                            } else if (tmp.equals("word2")) {
                                word2 = this.parseText();
                            } else if (tmp.equals("word3")) {
                                word3 = this.parseText();
                            } else {
                                this.parseIgnore();
                            }
                        }
                        // (from, subject, body);
                        for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
                            XmppListener xl = (XmppListener) e.nextElement();
                            xl.onThinkEvent(thinktype, from, pic, thinkword, word1, word2, word3);
                            //xl.onMessageEvent((from.indexOf('/') == -1) ? from : from.substring(0, from.indexOf('/')), body);
                        }
                                                /*
                                                 *  End of extension
                                                 */
                    } else {
                        this.parseIgnore();
                    }
                } else {
                    this.parseIgnore();
                }
            }
        }
    }
    
    /**
     * This method parses all presence stanzas, including subscription requests.
     *
     * @throws java.io.IOException is thrown if {@link XmlReader} or {@link XmlWriter}
     *	throw an IOException.
     */
    private void parsePresence() throws IOException {
        if (DEBUG) java.lang.System.out.println("*debug* parsing Presence");
        final String from = this.reader.getAttribute("from"), type = this.reader.getAttribute("type");
        String status = "", show = null;
        // int priority=-1;
        while (this.reader.next() == XmlReader.START_TAG) {
            final String tmp = this.reader.getName();
            if (tmp.equals("status")) {
                status = this.parseText();
            } else if (tmp.equals("show")) {
                show = this.parseText();
                // else if(tmp.equals("priority"))
                // priority = Integer.parseInt(parseText());
            } else {
                this.parseIgnore();
            }
        }
        
        if (DEBUG) java.lang.System.out.println("*debug* from=" + from);
        if (DEBUG) java.lang.System.out.println("*debug* type=" + type);
        if (DEBUG) java.lang.System.out.println("*debug* status=" + status);
        if (DEBUG) java.lang.System.out.println("*debug* show=" + show);
        
        if ((type != null) && (type.equals("unavailable") || type.equals("unsubscribed") || type.equals("error"))) {
            for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
                XmppListener xl = (XmppListener) e.nextElement();
                xl.onUnsubscribeEvent(from);
            }
        } else if ((type != null) && type.equals("subscribe")) {
            for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
                XmppListener xl = (XmppListener) e.nextElement();
                xl.onSubscribeEvent(from);
            }
        } else {
            final String jid = (from.indexOf('/') == -1) ? from : from.substring(0, from.indexOf('/'));
            for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
                XmppListener xl = (XmppListener) e.nextElement();
                xl.onStatusEvent(jid, show, status);
            }
        }
    }
    
    /**
     * This method parses all incoming messages.
     *
     * @throws java.io.IOException is thrown if {@link XmlReader} or {@link XmlWriter}
     *	throw an IOException.
     */
    private void parseMessage() throws IOException {
        if (DEBUG) java.lang.System.out.println("*debug* parsing Message");
        final String from = this.reader.getAttribute("from"), type = this.reader.getAttribute("type");
        String body = null, subject = null;
        while (this.reader.next() == XmlReader.START_TAG) {
            final String tmp = this.reader.getName();
            if (tmp.equals("body")) {
                body = this.parseText();
            } else if (tmp.equals("subject")) {
                subject = this.parseText();
            } else {
                this.parseIgnore();
            }
        }
        // (from, subject, body);
        for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
            XmppListener xl = (XmppListener) e.nextElement();
            xl.onMessageEvent((from.indexOf('/') == -1) ? from : from.substring(0, from.indexOf('/')), body);
        }
    }
    
    /**
     * This method parses all text inside of xml start and end tags.
     *
     * @throws java.io.IOException is thrown if {@link XmlReader} or {@link XmlWriter}
     *	throw an IOException.
     */
    private String parseText() throws IOException {
        if (DEBUG) java.lang.System.out.println("*debug* parsing Text");
        
        final String endTagName = this.reader.getName();
        final StringBuffer str = new StringBuffer("");
        int t = this.reader.next(); // omit start tag
        while (!endTagName.equals(this.reader.getName())) {
            if (t == XmlReader.TEXT) {
                str.append(this.reader.getText());
            }
            t = this.reader.next();
        }
        return str.toString();
    }
    
    /**
     * This method doesn't parse tags it only let the reader go through unknown
     * tags.
     *
     * @throws java.io.IOException is thrown if {@link XmlReader} or {@link XmlWriter}
     *	throw an IOException.
     */
    private void parseIgnore() throws IOException {
        if (DEBUG) java.lang.System.out.println("*debug* parsing Ignore");
        int x;
        while ((x = this.reader.next()) != XmlReader.END_TAG) {
            if (x == XmlReader.START_TAG) {
                this.parseIgnore();
            }
        }
    }
    
    /**
     * This method is used to be called on a parser or a connection error.
     * It tries to close the XML-Reader and XML-Writer one last time.
     *
     */
    private void connectionFailed() {
        try {
            this.writer.close();
            this.reader.close();
        } catch (final Exception e) {
            // io.printStackTrace();
        }
        for (Enumeration e = listeners.elements(); e.hasMoreElements();) {
            XmppListener xl = (XmppListener) e.nextElement();
            xl.onConnFailed();
        }
    }
    
    /**
     * The following methods are specifically developed for the ThinkAlike application
     * to execute the nescessary commands between the applications
     *
     * @athour Runar Gunnerud
     */
    
    /**
     * This methods sends a ThinkRequest from the initiator to the recipient
     * @param to the address of the recipient
     * @param from the address of the initiator
     * @param pic the picture represented as a Base64 encoded string
     * @param thinkword TheWord
     * @throws java.io.IOException in case of sending failure
     */
    public void sendThinkRequest(String to, String from, String pic, String thinkword) throws IOException {
        this.writer.startTag("iq");
        this.writer.attribute("id", "thinkalike");
        this.writer.attribute("type", "get");
        this.writer.attribute("to",to+"/thinkalike"); // add the resource of the recepient, always needed when sending an IQ message
        this.writer.attribute("from",from);
        this.writer.startTag("query");
        this.writer.attribute("xmlns", "com.runarg.thinkalike");
        
        this.writer.startTag("thinktype");
        this.writer.text("thinkrequest");
        this.writer.endTag(); // thinktype
        
        this.writer.startTag("pic");
        this.writer.text(pic);
        this.writer.endTag(); // pic
        
        this.writer.startTag("thinkword");
        this.writer.text(thinkword);
        this.writer.endTag(); // thinkword
        
        this.writer.endTag(); // query
        this.writer.endTag(); // iq
        this.writer.flush();
    }
    
    /**
     * This message is used by the recipient to send a response to the initiator of the ThinkRequest
     * @param to address of initiator
     * @param from address of recipiont
     * @param word1 .
     * @param word2 .
     * @param word3 .
     * @throws java.io.IOException on problems sending
     */
    public void sendThinkRespond(String to, String from, String word1, String word2, String word3) throws IOException {
        this.writer.startTag("iq");
        this.writer.attribute("id", "thinkalike");
        this.writer.attribute("type", "set");
        this.writer.attribute("to",to); // do not need to add resource in this case because it is send with the request massage and reused.
        this.writer.attribute("from",from);
        this.writer.startTag("query");
        this.writer.attribute("xmlns", "com.runarg.thinkalike");
        
        this.writer.startTag("thinktype");
        this.writer.text("thinkresponse");
        this.writer.endTag(); // thinktype
        
        this.writer.startTag("word1");
        this.writer.text(word1);
        this.writer.endTag(); // word1
        
        this.writer.startTag("word2");
        this.writer.text(word2);
        this.writer.endTag(); // word2
        
        this.writer.startTag("word3");
        this.writer.text(word3);
        this.writer.endTag(); // word3
        
        this.writer.endTag(); // query
        this.writer.endTag(); // iq
        this.writer.flush();
    }
};






